#require(MASS) 
#require(MCMCpack) 

# qcallrv=simple_call(data_matrix=data_matrix, arguments = arguments)
simple_call=function(data_matrix=NULL, data_array=NULL, data_and_details=NULL, simpleArguments=set_arguments(), 
                      preLDFquick=FALSE, quick=FALSE, vquick=FALSE, chr="NA", cnv_name="myCNV", pos="NA",
                      call_thresh=0.9, plotm=TRUE, plotum=TRUE, useLdf=FALSE)
{
  #You might want to do rough and dirty calls to seed the ldf and then do ldf more carefully
  checks = runInitialArgChecks(data_matrix=data_matrix, data_array=data_array, data_and_details=data_and_details, arguments=simpleArguments)
  if(preLDFquick) nonLdfQuick=T else nonLdfQuick=quick
  rv             = analyse_data(dat_matrix=data_matrix, data_array=data_array, data_and_details=data_and_details, 
                    quick=nonLdfQuick, vquick=vquick, arguments=simpleArguments)
  if(useLdf)
  {
    rv$data_and_details$arguments = simpleArguments
    ldf_data_and_details = make_ldf_data_and_details(data_and_details=rv$data_and_details, first_round_results = rv)
    rv                   = analyse_data(arguments=simpleArguments, extra_args=NULL, data_and_details=ldf_data_and_details,
                             DEBUG=ldf_data_and_details$arguments$DEBUG, quick=quick, vquick=vquick)
  }
  noco           = rv$best_fit$opars$noco
  collnames      = rownames(rv$best_fit$opars$hkclm)
  unmerged_nocl  = rv$best_fit$opars$nocl
  calls_matrices = get_calls_matrices(objects=rv$best_fit, ung_dat=rv$ung_dat, collnames = names(data_array))
  chiamo_style   = as.list(rep(NA, noco)); merged_calls = as.list(rep(NA, noco)); merged_hcalls = as.list(rep(NA, noco));  unmerged_hcalls = as.list(rep(NA, noco))
  names(chiamo_style)=collnames; names(merged_calls)=collnames; names(merged_hcalls)=collnames;  names(unmerged_hcalls)=collnames; 
  for(coll in 1:noco)
  {
   #coll=1; cat("  Warning, for now just doing one collection.\n");
    chiamo_style[[coll]]    = matrix(c(chr, cnv_name, pos, unmerged_nocl, round(calls_matrices[[coll]], digits=3)), nrow=1)
    unmerged_hcalls[[coll]] = make_hcalls(calls_mat=t(calls_matrices[[coll]]), call_thresh=call_thresh)
    merged_fit              = post_bic_simple_merge(fit=rv$best_fit, pars=rv_best_fit$opars, mode_counter="simple_nc")
    merged_calls[[coll]]    = merge_calls(calls=chiamo_style[[coll]], relabelled_classes=merged_fit$relabelled_classes, return_mat=T, ind_names=colnames(calls_matrices[[coll]]));
    merged_hcalls[[coll]]   = make_hcalls(calls_mat=merged_calls[[coll]], call_thresh=call_thresh)
  }
  if(plotm | plotum) par(mfcol=c(plotm+plotum,1))
  if(plotm)  plotd(objects=merged_fit, dens_vals = merged_fit$mdens_vals, model_fits = merged_fit$merged_fits, main="merged", plot_combined=F, new_plot=F)
  if(plotum) plotd(objects=rv$best_fit, main="unmerged", plot_combined=F, new_plot=F)
  return_val =      list(rv$best_fit , rv$stpt_states  ,  rv$cnv_name   ,  calls_matrices ,  unmerged_hcalls ,  merged_fit ,
                        rv$data_array, rv$sample_names ,  rv$ung_dat    ,  merged_calls   ,  merged_hcalls   ,  rv$kernel_mfits )
  names(return_val) = c(   "best_fit",   "stpt_states" ,    "cnv_name"  , "calls_matrices", "unmerged_hcalls", "merged_fit",
                         "data_array",   "sample_names",    "ung_dat"   , "merged_calls"  , "merged_hcalls"  , "kernel_mfits")
  return(return_val)  
}


call_cnvs=function(arguments, in_data_and_details=NULL)
{
  if(!exists("DEBUG")) {cat("Using DEBUG from arguments object\n"); DEBUG=arguments$DEBUG} else arguments$DEBUG = DEBUG
  extra_args         = set_up_caller(arguments)
  for(cnv_index in arguments$cnvs) {
    if(is.null(in_data_and_details)) {
      data_and_details = get_data(      arguments=arguments, extra_args=extra_args, cnv_index=cnv_index, DEBUG=DEBUG);
    } else data_and_details = in_data_and_details
    abort = data_and_details$abort; 
    if(abort) {fit="aborted_run"} else { #break();
      if(!arguments$write_sim_to_file) {
        arguments        = data_and_details$arguments; if(data_and_details$collect_garbage) gc(verbose=FALSE)
        results          = analyse_data(  arguments=arguments, extra_args=extra_args, data_and_details=data_and_details, DEBUG=DEBUG)
        fit              = output_results(arguments=arguments, extra_args=extra_args, data_and_details=data_and_details, results=results)
        ########### One way to run LDF here #############
        if(arguments$make_ldf_calls)
        {
          ldf_data_and_details = make_ldf_data_and_details(data_and_details    =data_and_details, first_round_results = results)
         #ldf_results          = run_ldf_with_main_routine(data_and_details=ldf_data_and_details, first_round_results = results)
          ldf_results          = analyse_data(arguments=arguments, extra_args=extra_args, data_and_details=ldf_data_and_details,
                                   DEBUG=ldf_data_and_details$arguments$DEBUG)
          ldf_best_fit         = output_ldfCalls(arguments=arguments, data_and_details=ldf_data_and_details, ldf_results=ldf_results)
        }
      }
    }
  }
  if(exists("ldf_best_fit")) return(ldf_best_fit) else {
    if(exists("fit")) return(fit) else {
      cat("  Error, object fit hasn't been made.\n"); cat("On cnv: ", cnv_index, " from total of ", arguments$cnvs, "\n", sep="")
      if(DEBUG > 1.7) browser();
      return(NULL)
    }
  }
}

#run_ldf_with_main_routine=function(data_and_details)
#{
#  ldf_results = analyse_data(arguments=arguments, extra_args=extra_args, data_and_details=data_and_details, DEBUG=data_and_details$arguments$DEBUG)
#  return(ldf_results)
#}

make_ldf_data_and_details=function(first_round_results, data_and_details, wdat=0.1)
{
  dat_matrix        = data_array_to_matrix(data_array=data_and_details$data_array)
  cat("  Calcualating ldf probe-summary...");
  calls_matrices    = get_calls_matrices(objects=first_round_results$best_fit, 
                                         ung_dat=first_round_results$ung_dat , 
                                         collnames = names(data_and_details$data_array))
  # Changed sapply to lapply on this line as sapply wasn't returning a list when nocl==1
  calls_mat         = data_array_to_matrix(data_array = lapply(calls_matrices, t))
  ldf_summary       = perform_ldf(calls_mat=calls_mat, dat_matrix=dat_matrix)
  cat("  Done\n");
  ldf_mat           = matrix(ldf_summary, ncol=1)
  colnames(ldf_mat) = "all_colls"
  rownames(ldf_mat) = rownames(calls_mat)
  data_and_details2 = data_and_details
  data_array        = data_matrix_to_array(data_matrix=ldf_mat)

  # BUG: data_array no longer contains individual probes here, so PVS scaling
  # screws up.  Change to no probe scaling as a temporary fix.  This may
  # actually be the correct thing to do -- PVS scaling might make no sense at
  # all when doing LDA.  Need to check this.  If it *does* make sense, it will
  # need to be implemented before the actual LDA step (i.e. somewhere earlier
  # than here).
  no_inds_vec = unlist(apply(data_array, 2, get_no_inds))
  fdat        = process_data(data_array,
                             no_inds          = sum(no_inds_vec),
                             keep_range       = FALSE,
                             wdat             = wdat,
                             DEBUG            = data_and_details$arguments$DEBUG,
                             return_ungridded = TRUE,
                             grid_density     = data_and_details$arguments$grid_density,
                             probe_scaling    = "none",
                             probe_summary    = data_and_details$arguments$probe_summary,
                             verbose          = data_and_details$arguments$verbose)
  data_and_details2$data_array   = data_array
  data_and_details2$gdat         = fdat$gridded
  data_and_details2$ung_dat      = fdat$ungridded
  data_and_details2$grid_breaks  = fdat$grid_breaks
  data_and_details2$true_d_range = fdat$true_d_range  
  data_and_details2$probe_scale  = fdat$probe_scale  # the option used for probe scaling
  return(data_and_details2)
}



# source("cnv_caller.R"); mat = matrix(rnorm(1e4, 0, 1), nrow=10, ncol=1000); mat[,501:1000] = mat[,501:1000] + 2.3; rv = analyse_data(dat_matrix=mat, DEBUG=1.71); fit = rv$best_fit; plotd(fit, no_bars=30, new_plot=T)
# source("cnv_caller.R"); mat = matrix(rnorm(3e4, 0, 1), nrow=10, ncol=3000); mat[,501:1000] = mat[,501:1000] + 2.3; mat[,1001:2000] = mat[,1001:2000] + 1.1; mat[,2001:3000] = mat[,2001:3000] + 0.6; rv = analyse_data(dat_matrix=mat, DEBUG=1.71, nocls=10:6); fit = rv$best_fit; plotd(fit, no_bars=30, new_plot=T)
analyse_data=function(arguments=NULL, extra_args=NULL, data_and_details=NULL, data_array=NULL, dat_matrix=NULL, 
                      no_iters=NULL, precision=1e-5, DEBUG=1.71, nocls=NULL, quick=F, vquick=F)
{
  if(!exists("DEBUG", envir=.GlobalEnv)) assign(x="DEBUG", value = DEBUG, env=.GlobalEnv)
  if(is.null(data_and_details) & is.null(data_array) & is.null(dat_matrix)) {
    cat("  Error - must specify either: data_and_details, data_array or dat_matrix\n  Exiting.\n");
    return(1)
  }
  if(is.null(data_and_details) & is.null(data_array) & is.null(dat_matrix)) {
    cat("  Warning - no individual names found. Labelling individual according to rownumber.\n");
    rownames(dat_matrix) = paste("ind" , 1:nrow(dat_matrix), sep="")
    colnames(dat_matrix) = paste("prob", 1:ncol(dat_matrix), sep="")
  }
  if(is.null(arguments))        {
    arguments          = set_arguments(write_results=F, quick=quick, vquick=vquick, nocls=nocls, cnvs=1)
  }
  if(is.null(extra_args))        {
    extra_args         = set_up_caller(arguments)
  }
  if(!is.null(no_iters ))        arguments$no_iters = no_iters
  if( is.null(data_and_details)) {
    data_and_details   = make_data_and_details(dat_matrix=dat_matrix, data_array=data_array, arguments=arguments)
    if(length(arguments$case_colls)!= ncol(data_and_details$data_array)) arguments$case_colls = makeCaseColls(noco=ncol(data_and_details$data_array))
  }
  
  no_iters  = arguments$no_iters;  
  noco      = data_and_details$noco;                
  cnv_name  = data_and_details$cnv_name;
  cnv_index = data_and_details$cnv_index
  nocls     = arguments$nocls
  sa_limit  = arguments$set_sa_limit*no_iters
  stpts     = extra_args$stpts
  ## This is a debugging tool - to allow easy debugging on a certain iteration where something might be going wrong.
  artificial_stop = c(arguments$artificial_stop_i, arguments$artificial_stop_cl, arguments$artificial_stop_cn, arguments$artificial_stop_st)

  ## This is a non-standard way tto run the function - as this is part of code that either grabs or otherwise processes the data this can be used
  ## To examine that data visually (and do no inference)
  if(arguments$visualise) {
    visualise_data(data_array=data_and_details$data_array, ung_dat=data_and_details$ung_dat, g_dat=data_and_details$dat, noco=noco, cnv_name=cnv_name, cnv_index=cnv_index)
    return(NULL);
  }

  #Debug-code: Objects to record the various model fits and states throughout fitting
  stpt_states           = matrix(list(), nrow=length(nocls), ncol=length(stpts))
  rownames(stpt_states) = paste(rep("cl=", length(nocls)), nocls, sep="")
  if(is.null(names(stpts))) names(stpts) = paste("stpt", 1:length(stpts), sep="")
  colnames(stpt_states) = names(stpts)
  acpt_or_rjct          = array(-1, c(1, no_iters, length(nocls)))
  model_likelihoods           = matrix(-1E200, nrow=length(nocls), ncol=length(stpts))
  rownames(model_likelihoods) = paste(nocls, rep("_clusters", length(nocls)), sep="")
  colnames(model_likelihoods) = names(stpts)
  accept   = matrix(-1, nrow=length(nocls), ncol=no_iters)
  mv_names = matrix(-1, nrow=length(nocls), ncol=no_iters)

  if(arguments$use_hier_mchoice) mchoice_set_point = "hier_mchoice" else mchoice_set_point = "simple"
  if(DEBUG>0 & arguments$verbose>=1 & !arguments$real_data) cat(" Looking at cnv ", cnv_index, "\n", sep="");
  for(nocl in nocls)
  {
    #the index for the number of classes being looked at
    nocli                = which(nocls==nocl)
    if(nocl>1) stpt_inds = 1:length(stpts) else stpt_inds = 1:length(stpts) #Used to just have one start-point for 1-class as they should not be very different
    no_stpts             = length(stpt_inds)
    last_nocl_evs        = sapply(X=stpt_states[(nocli-1),], FUN=calc_evidence, separate_collections=arguments$use_hier_mchoice, set_point = mchoice_set_point, type="bic") #objects$lstore$hier_mchoice$coll_spec_liklihd
    for(stpt_ind in stpt_inds) # stpt_ind is short for 'start point index' the model tries starting from various 'places in the space'
    {
      if(DEBUG>0 & stpt_ind==1 & arguments$verbose>=1) cat(" Looking at using ", nocl, " classes.\n", sep="");
      if(DEBUG==3.111) {cat("In run_caller (stopped as DEBUG=3.111), about to initialise data\n"); browser();}
      #*************************************#
      objects = Initialise(nocl=nocl, stpt_ind=stpt_ind, cnv_index=cnv_index, extra_args=extra_args, arguments=arguments, data_and_details=data_and_details, model="simple", DEBUG=DEBUG)
      #*************************************#
      inherit_inds  = (no_stpts:(no_stpts-(nocl+1)))[1:min(no_stpts, nocl+1)]
      if(arguments$inherit_stps & nocl!=max(nocls)) {
        if(stpt_ind%in%inherit_inds) {
          del_class = which(inherit_inds==stpt_ind) 
        }  else del_class=NA
        if(!is.na(del_class)) {
          objects$flags$inherit_stpt = T 
          if(arguments$use_hier_mchoice) model="hier_nocl_choice" else model="simple"
          objects$flags$iter = max(objects$flags$simple_model_iters)
          iters_so_far       = objects$flags$iter
          objects$opars      = make_new_start_pars(  stpt_states[nocli-1, which.max(last_nocl_evs)][[1]]$opars  , del_class=del_class, model_pdisease=objects$flags$model_pdisease)
          objects$opriors    = make_new_start_priors(stpt_states[nocli-1, which.max(last_nocl_evs)][[1]]$opriors, del_class=del_class)
          objects$npars      = objects$opars; objects$npriors = objects$opriors;
          objects            = recalc_likelihood( objects) 
          objects            = update_olikelihood(objects=objects, model=model)
          objects$lstore$simple_model = NA
          bad     = check_clm(objects=objects, move="update_params", cur=T); if(bad) {cat("In run_caller. Something wrong with parameter inheritence.\n"); browser();}
         #cat("Check that I'm making some sense at all with using previous model fit to inform new start_points!\n"); browser();
        }
        #if(cnv_index==2 & nocls[nocli]==1) browser();
      } 
      if(!arguments$inherit_stps | nocl==max(nocls) | !stpt_ind%in%inherit_inds | !arguments$use_hier_mchoice) {
        iters = next_iters(iters_so_far=0, extra_iters = arguments$simple_model_iters)
        rv = perform_these_updates(objects=objects, iters = iters, acpt_or_rjct=acpt_or_rjct, mv_names=mv_names, nocli=nocli, 
                DEBUG=arguments$DEBUG, model="simple", arguments=arguments)
        objects                           = rv$objects; 
        objects$lstore$simple_mchoice = objects$likelihood; 
        objects$ninferred$simple_pars = objects$opars; objects$oinferred$simple_pars = objects$opars; 
        acpt_or_rjct = rv$acpt_or_rjct; mv_names = rv$mv_names;
        check_simple_model_results(objects)      
        iters_so_far  = rv$iter; 
      }
      if(arguments$use_hier_mchoice) {
        iters = next_iters(iters_so_far=iters_so_far, extra_iters = arguments$hier_mchoice_iters)
        rv = perform_these_updates(objects=objects, iters = iters, acpt_or_rjct=rv$acpt_or_rjct, mv_names=rv$mv_names, nocli=nocli, 
                DEBUG=arguments$DEBUG, model="hier_nocl_choice", arguments=arguments)
        objects = rv$objects
        objects$likelihood$coll_spec_liklihds = calc_intens_lkld(dat=objects$dat, best_call=!objects$flags$sum_calls, test=F, objects=objects, new=F, colls_separate=T)
        objects$lstore$hier_mchoice           = objects$likelihood 
        objects$ninferred$hier_mchce_pars     = objects$opars; objects$oinferred$hier_mchce_pars = objects$opars; 
        iters_so_far  = rv$iter; 
      }
     #model_likelihoods[nocli, stpt_ind]  = objects$likelihood$simple_liklihd;
      stpt_states[nocli,stpt_ind][[1]]    = objects
      if(DEBUG>0.5) if(length(stpt_states[nocli,stpt_ind][[1]])!=12){cat("Looks like the structure of stpt_states has gone wrong.\n"); browser();}
    }
    #for(i in 1:min(nocl, ncol(stpt_states)))
    #{
    #  stpt_ind                                              = ncol(stpt_states)-i+1
    #  stpt_states[nocli-1, stpt_ind]$flags$inherit_stpt     = T 
    #  stpt_states[nocli-1, stpt_ind]$oinferred$start_pars   = make_new_start_pars(  stpt_states[nocli, which.max(this_nocl_evs)]$opars  , del_class=i)
    #  stpt_states[nocli-1, stpt_ind]$oinferred$start_priors = make_new_start_priors(stpt_states[nocli, which.max(this_nocl_evs)]$opriors, del_class=i)
    #}
  }
  ## (Number of Classes) MODEL CHOICE DONE HERE.
  #  cat("done, (took", proc.time()-ptm, "seconds)\n"); browser();
  #  evidences = calc_evidence(fit=rv$objects, separate_collections=T, set_point = "hier_mchoice") #objects$lstore$hier_mchoice$coll_spec_liklihd
  #  vcbe = Vectorize(calc_evidence, vectorize.args=c("separate_collections", "set_point"))
  #  evs = vcbe(stpt_states, separate_collections=T, set_point = "hier_mchoice") #Doesn't work this way. Passes whole of stpt_states!
  #  evs2 = vcbe(as.vector(stpt_states[2,]), separate_collections=T, set_point = "hier_mchoice") #Doesn't work this way. Passes whole object!
  #checksame# e2 = apply(X=stpt_states, MARGIN=c(1,2), FUN=calc_evidence, separate_collections=arguments$use_hier_mchoice, set_point = NULL, type="bic") 
  evidences      = apply(X=stpt_states, MARGIN=c(1,2), FUN=calc_evidence, separate_collections=arguments$use_hier_mchoice, set_point = mchoice_set_point, type="bic") 
  aic_evidences  = apply(X=stpt_states, MARGIN=c(1,2), FUN=calc_evidence, separate_collections=arguments$use_hier_mchoice, set_point = mchoice_set_point, type="aic") 

  if(is.null(arguments$use_supplied_nocl)) choose_nocl_from_fit = T else {
    if(arguments$use_supplied_nocl=="none") choose_nocl_from_fit = T else choose_nocl_from_fit = F
  }
  ## New code as of Fri17thJuly09, Added when I realised that model-choice was bad with (many individuals and) normal data with t8 kernels
  kernel_mfits = NULL
  if(arguments$infer_kernel) {
    if(DEBUG>0 & arguments$verbose>=1) cat(" Inferring what kernel to use.\n", sep="");
    best_stpts_colls = apply(X=evidences, MARGIN=1, FUN=which.max)
    if(DEBUG>=2.0 & arguments$verbose>=1) {cat("  Evidences before updating signal models.\n"); print(apply(X=evidences, MARGIN=1, FUN=max))}
    no_kernels       = length(arguments$possible_kernels)
    kernel_names     = paste(names(arguments$possible_kernels), arguments$possible_kernels, sep="_")
    check_noke       = length(no_kernels>1); if(!check_noke & DEBUG>0.1) {cat("  Error, exploring possible kernels yet don't have multiple possibilities.\n"); if(DEBUG>=1) browser()}
    kernel_mfits           = matrix(list(), nrow=length(nocls), ncol=no_kernels)
    rownames(kernel_mfits) = rownames(stpt_states)
    colnames(kernel_mfits) = kernel_names
    kernel_evidences       = matrix(NA, nrow=length(nocls), ncol=no_kernels)
    dimnames(kernel_evidences) = dimnames(kernel_mfits)
    # This vector records the model fits for the best number of classes for each kernel
    best_kernel_fits       = as.list(rep(NA, no_kernels));    names(best_kernel_fits) = kernel_names
    kernel_indices   = 1:no_kernels
    for(kernel_index in kernel_indices)
    {
      kernel      = names(arguments$possible_kernels[kernel_index]) # This means either 't' or 'normal'
      kernel_name = kernel_names[kernel_index] # This is a way to access data-structures referring to this specific kernel: eg t_3, t_8 or normal_NA (kernel-type_degrees-freedom)
      if(DEBUG>=1 & arguments$verbose>1) cat(" On ", kernel_name, ".\n", sep="");
      deg_free    = arguments$possible_kernels[kernel_index]
      for(nocl in nocls)
      {
        nocli        = which(nocls==nocl)
        nocl_rowname = paste("cl=", nocl, sep="")
        stpt_row     = which(rownames(stpt_states)==nocl_rowname)
        objects      = stpt_states[[stpt_row, best_stpts_colls[stpt_row]]]
        objects$flags$signal_model = kernel
        objects$flags$deg_free     = deg_free
        objects      = recalc_likelihood( objects) 
        objects      = update_olikelihood(objects=objects, model="no_assoc")
        iters_so_far = objects$flags$iter
        iters        = next_iters(iters_so_far=iters_so_far, extra_iters = arguments$kernel_choice_iters)
        # The 'model' parameter here controls: Whether beta_1 can be changed, and whether there's a hierarchy, and therefore how to calculate the model evidence.
        rv                                    = perform_these_updates(objects=objects, iters = iters, nocli=NA, 
                                                  DEBUG=arguments$DEBUG, model="hier_nocl_choice", arguments=arguments)
        objects                               = rv$objects
        objects$likelihood$coll_spec_liklihds = calc_intens_lkld(dat=objects$dat, best_call=!objects$flags$sum_calls, test=F, objects=objects, new=F, colls_separate=T)
        objects$lstore$hier_mchoice           = objects$likelihood 
        objects$ninferred$hier_mchce_pars     = objects$opars; objects$oinferred$hier_mchce_pars = objects$opars; 
        kernel_mfits[[nocl_rowname, kernel_name]]   = objects
        kernel_evidences[nocl_rowname, kernel_name] = calc_evidence(kernel_mfits[[nocl_rowname, kernel_name]], separate_collections=TRUE, set_point=NULL, type="bic")
        iters_so_far  = rv$iter; 
      }
      best_kernel_fit_indices         = which.max(kernel_evidences[,kernel_name])
      best_kernel_fits[[kernel_name]] = kernel_mfits[[best_kernel_fit_indices,kernel_name]]
    }
    best_kernel_index = which(kernel_evidences==max(kernel_evidences), arr.ind=T)[1,,drop=T]
    names(best_kernel_index) = c("nocl_index", "kernel")
    best_fit          = kernel_mfits[[best_kernel_index["nocl_index"], best_kernel_index["kernel"]]]
    if(DEBUG>1 & arguments$verbose>= 1) {
      #cat(" Chosen kernel: ", colnames(kernel_evidences)[best_kernel_index["kernel"]], "\n", sep="");
      cat("  Chosen signal model =", best_fit$flags$signal_model)
      if(best_fit$flags$signal_model=="t") cat(", chosen degrees of freedom =", best_fit$flags$deg_free)
      cat("\n")
    }
    if(DEBUG>2.0 & arguments$verbose>=1) {cat("  Evidences after updating signal models.\n"); print(kernel_evidences)}
  } else {
    best_model     = which(    evidences==max(    evidences, na.rm=T), arr.ind=T)
    aic_bestml     = which(aic_evidences==max(aic_evidences, na.rm=T), arr.ind=T)
    aic_bestnocl   = nocls[aic_bestml[1]]
    if(DEBUG>0 & arguments$verbose > 0) {cat(" bic best nocl = ", nocls[best_model[1]], ", and aic best nocl = ", nocls[aic_bestml[1]], "\n", sep="")}
    if(nocls[aic_bestml[1]]<nocls[best_model[1]]) {cat(" incompatible aic vs bic model choice.\n"); browser();}
    if(choose_nocl_from_fit)
      best_fit            = stpt_states[best_model][[1]]
    if(!choose_nocl_from_fit)
    {
      sdvs = get_cnv_verdicts(which_verdicts=arguments$use_supplied_nocl, verdict_col=arguments$supplied_nocl_col, simplify=T)
      verdicts_row        = which(supplied_nocls[,"CNV"]==cnv_name)
      if(length(verdicts_row)!=1) {
        cat("  Warning!! Couldn't find this CNV in the verdicts file. Choosing number of classes using BIC.\n");
        best_fit            = stpt_states[best_model][[1]]
      }
    }
  }

  if(is.null(arguments$use_supplied_nocl)) choose_nocl_from_fit = T else {
    if(arguments$use_supplied_nocl=="none") choose_nocl_from_fit = T else choose_nocl_from_fit = F
  }
  if(!choose_nocl_from_fit)
  {
    sdvs                = get_cnv_verdicts(which_verdicts=arguments$use_supplied_nocl, verdict_col=arguments$supplied_nocl_col, simplify=T)
    verdicts_row        = which(supplied_nocls[,"CNV"]==cnv_name)
    this_nocl           = sdvs[verdicts_row]
    if(arguments$infer_kernel) {
      nocl_row            = which(rownames(kernel_mfits)==paste("cl=", this_nocl, sep=""))
      best_for_this_nocl  = which.max(kernel_evidences[nocl_row,])
      best_fit            = kernel_mfits[nocl_row, best_for_this_nocl][[1]]
    } else 
    {
      nocl_row            = which(rownames(stpt_states)==paste("cl=", this_nocl, sep=""))
      best_for_this_nocl  = which.max(evidences[nocl_row,])
      best_fit            = stpt_states[nocl_row, best_for_this_nocl][[1]]
    }
  }
  if(DEBUG>0 & arguments$verbose>=1) cat(" Extra iterations for chosen number of classes (", best_fit$opars$nocl, ").\n", sep="");
  if(best_fit$flags$sum_calls_assoc!=best_fit$flags$sum_calls_nocl) {
     best_fit$opars            = best_fit$npars #Shouldn't actually be necessary, but worth making sure!
     best_fit$flags$sum_calls  = best_fit$flags$sum_calls_assoc
     best_fit                  = recalc_likelihood(objects=best_fit) 
    #best_fit                  = reset_nlikelihood(objects = best_fit)
  }
  ## Assocation testing and full model fitting once the number of classes is chosen (also, of course, the start point). Can run a lot more here!  
  best_accepts  = rep(NA, max(arguments$simple_model_iters)+max(arguments$no_assoc_model_iters)+max(arguments$full_model_iters))
  best_mv_names = rep(NA, max(arguments$simple_model_iters)+max(arguments$no_assoc_model_iters)+max(arguments$full_model_iters))
  iters_so_far  = rv$iter
  #### NO ASSOCATION MODEL iterations ####
 #iters = (iters_so_far+1):(iters_so_far+max(arguments$no_assoc_model_iters))
  iters = next_iters(iters_so_far=iters_so_far, extra_iters = arguments$no_assoc_model_iters)

  rv = perform_these_updates(objects=best_fit, iters = iters, best_acpt_or_rjct=best_accepts, best_mv_names=best_mv_names, 
               nocli=nocli, DEBUG=arguments$DEBUG, model="no_assoc", arguments=arguments)
  best_fit      = rv$objects
  best_accepts  = rv$best_acpt_or_rjct
  best_mv_names = rv$best_mv_names
  best_fit$lstore$no_assoc = best_fit$likelihood
  best_fit$ninferred$noassoc_pars = rv$objects$opars; best_fit$oinferred$noassoc_pars = rv$objects$opars; 

 #debug(calc_phen_g_gen); x = calc_phen_g_gen(objects=best_fit, hard_calls=!objects$flags$sum_calls)
  #### ADDITIVE ASSOCIATION MODEL iterations ####
 #iters = (iters_so_far+1):(iters_so_far+max(arguments$full_model_iters))
  iters = next_iters(iters_so_far=iters_so_far, extra_iters = arguments$full_model_iters)
  if( arguments$use_hierarchy) final_model = "full_model"
  if(!arguments$use_hierarchy) final_model = "full_no_hierarchy"
  rv = perform_these_updates(objects=best_fit, iters = iters, best_acpt_or_rjct=best_accepts, best_mv_names=best_mv_names, 
               nocli=nocli, DEBUG=arguments$DEBUG, model=final_model, arguments=arguments)
 #if(final_model=="full_no_hierarchy")
  best_fit      = rv$objects
  best_accepts  = rv$best_acpt_or_rjct
  best_mv_names = rv$best_mv_names
  best_fit$lstore$full_model = best_fit$likelihood
  best_fit$likelihood$full_liklihd = best_fit$likelihood$o_liklihd
  if(arguments$use_hier_mchoice) {
    best_fit$likelihood$bic_evidence = best_fit$lstore$hier_mchoice$bic_evidence
    best_fit$likelihood$aic_evidence = best_fit$lstore$hier_mchoice$aic_evidence
  } else {
    best_fit$likelihood$bic_evidence = best_fit$lstore$simple_mchoice$bic_evidence
    best_fit$likelihood$aic_evidence = best_fit$lstore$simple_mchoice$aic_evidence
  }

 #debug(test_for_association)
  best_fit$likelihood$lassoc_bf = test_for_association(additive_assoc_lklihd = best_fit$lstore$full_model$o_liklihd, no_data_points = sum(best_fit$dat),
                                                    no_assoc_lklihd = best_fit$lstore$no_assoc$o_liklihd)
  
  if(DEBUG>1.5 & arguments$verbose>=1 & best_fit$likelihood$lassoc_bf>4 & best_fit$likelihood$lassoc_bf<18 )  cat(" Pretty strong evidence of association at this CNV (", best_fit$likelihood$lassoc_bf, ").\n", sep="")
  if(DEBUG>1.5 & arguments$verbose>=1 & best_fit$likelihood$lassoc_bf>18& best_fit$likelihood$lassoc_bf<500)  cat(" VERY strong evidence of association at this CNV.("  , best_fit$likelihood$lassoc_bf, ").\n", sep="")
  if(DEBUG>1.5 & arguments$verbose>=1 & best_fit$likelihood$lassoc_bf>500                                  ){ cat(" ERROR: STUPIDLY strong evidence of association at this CNV.("  , best_fit$likelihood$lassoc_bf, ").\n", sep=""); if(DEBUG > 1.7) browser()}

  if(arguments$store_stpts) {
    main_dir   = get_coutput_dirs()$coutput_root
    stpts_dir  = get_coutput_dirs()$stpts_dir
    if(!file.exists(main_dir )) dir.create(main_dir , showWarnings = FALSE)
    if(!file.exists(stpts_dir)) dir.create(stpts_dir, showWarnings = FALSE)

    if(is.null(cnv_name)) fn_cnv_name=cnv_index else fn_cnv_name=cnv_name;
    stpts_fn   = get_coutput_fns(cnv_name = fn_cnv_name)$stpts_fn
    model_fits =      list(stpt_states ,  best_fit )
    names(model_fits) = c("stpt_states", "best_fit")
    save(model_fits, file=stpts_fn)
  } else stpts_fn="none"
  if(0 & best_fit$opars$betas[2]==0 & arguments$DEBUG>1.71 & length(arguments$full_model_iters)>50) { 
    cat("  How have the betas managed to change by nothing?\n"); 
    browser();
  }
  return_val =      list(best_fit ,  model_likelihoods ,  stpt_states , "stpts_fn", data_and_details$cnv_name ,  kernel_mfits ,  data_and_details , 
      data_and_details$data_array ,  data_and_details$sample_names ,  arguments$cnv_types ,  data_and_details$ung_dat ,  data_and_details$total_sim )
  names(return_val) = c("best_fit", "model_likelihoods", "stpt_states",  stpts_fn ,                 "cnv_name", "kernel_mfits", "data_and_details",
                      "data_array",                  "sample_names",           "cnv_types",                  "ung_dat",                  "total_sim")
  return(return_val)
}

# iters = next_iters(iters_so_far=iters_so_far, extra_iters = arguments$hier_mchoice_iters)
next_iters=function(iters_so_far, extra_iters)
{
  if(is.null(iters_so_far)) iters_so_far=0 #Done this way to cope with null or 0 iters_so_far. 
  if(iters_so_far==0) {
    iters = extra_iters
    if(length(iters)==1)
      if(iters==0) iters=NULL
    return(iters)
  }
  if(max(extra_iters)> 0)
    iters = (iters_so_far+1):(iters_so_far+max(extra_iters))
  if(max(extra_iters)==0)
    iters = NULL
  return(iters)
}

perform_these_updates=function(objects, iters, acpt_or_rjct=NULL, mv_names=NULL, best_acpt_or_rjct=NULL, best_mv_names=NULL, 
                               nocli, DEBUG=1, model, arguments)
{
   objects$flags$model     = model; vb = arguments$verbose
   if(model %in% c("full_model", "full_no_hierarchy", "no_assoc")) objects$flags$sum_calls = objects$flags$sum_calls_assoc else objects$flags$sum_calls = objects$flags$sum_calls_nocl
  #if(model=="full_model") {debug(choose_move); browser()}
   if(DEBUG>1 & arguments$verbose>=2) cat("On iter: ")
   for(iter in iters)
   { 
     objects$flags$iter = iter
     objects = update_params(objects, iter=iter, extra_args=extra_args, arguments=arguments, model=model, last_iter=max(iters), DEBUG=DEBUG)
    #o = update_params(objects=best_fit, iter=10, extra_args=extra_args, arguments=arguments, model="full_model", last_iter=10)
     if(DEBUG>=0.5 &vb>0)          if(iter/(1000/vb) ==floor(iter/(1000/vb))) cat(" Iteration ", iter, " (model: ", model, ")", "\n", sep="")
     if(!is.null(acpt_or_rjct))      acpt_or_rjct[1,iter,nocli] = objects$accepted_last;  
     if(!is.null(mv_names    ))      mv_names[nocli,iter]       = names(objects$accepted_last)
     if(!is.null(best_acpt_or_rjct)) best_acpt_or_rjct[iter]    = objects$accepted_last;  
     if(!is.null(best_mv_names    )) best_mv_names[iter]        = names(objects$accepted_last)
     if(DEBUG>1 & arguments$verbose>=2 & iter%%15==0) cat(" ", iter, sep="")
   }
   if(DEBUG>1 & arguments$verbose>=2) cat("\n")
   rv =     list( objects ,  acpt_or_rjct ,  mv_names ,  iter ,  best_mv_names ,  best_acpt_or_rjct )
   names(rv) = c("objects", "acpt_or_rjct", "mv_names", "iter", "best_mv_names", "best_acpt_or_rjct")
   return(rv)
}

# Warning - may slow things down? What's the cost of these packages?
#raven_source = "C:/Documents and Settings/cardin/My Documents/Programming/cnv_caller/"
options(error=recover)
read_resource_files=function(scripts_dir=getwd(), wd=NULL, cccPlusproj=FALSE)
{
  source(file.path(scripts_dir, "cnv_caller_core.R"        ));
  source(file.path(scripts_dir, "moves_core.R"             ));
  source(file.path(scripts_dir, "calc_probs_core.R"        ));
  source(file.path(scripts_dir, "update_core.R"            ));
  source(file.path(scripts_dir, "initialise_core.R"        ));
  source(file.path(scripts_dir, "fmoves_core.R"            ));
  source(file.path(scripts_dir, "diagnostics_core.R"       ))
  source(file.path(scripts_dir, "math_fns_core.R"          ))
  source(file.path(scripts_dir, "using_mclust_core.R"      ))
  source(file.path(scripts_dir, "display_core.R"           ))
  source(file.path(scripts_dir, "post_calling_core.R"      ))
  source(file.path(scripts_dir, "pre_proc_data_core.R"     ))
  source(file.path(scripts_dir, "file_handling_core.R"     ))

  if(cccPlusproj) {
    source(file.path(scripts_dir, "find_cccplus_files_proj.R"))
    source(file.path(scripts_dir, "run_cnvtest_proj.R"       ))
  }
}
#if(exists("project_root") & !exists("scripts_dir")) scripts_dir = file.path(project_root, "caller_scripts/")
#if(exists("scripts_dir")) read_resource_files(scripts_dir=scripts_dir) else read_resource_files()


check_simple_model_results=function(objects, precision=1e-10)
{
  ## The below line only works when o_intsty is calculated with all collections the same.
  #  o_intsty comes from the full model, simple_liklihd uses averages for cluster centres etc. (which is fine when they're all the same)
  if((sum(calc_simple_likelihood(objects)) - objects$likelihood$o_intsty)>precision) {
    cat("Warning. In analyse_data. Inconsistent values for the likelihood.\n"); 
    simple = make_combined_lpdga(objects=objects, vals=as.numeric(rownames(objects$dat)), use_null=objects$flags$use_null_class)
    full   = update_lpdga(objects = objects)$ninferred$lpdga
    par(mfcol=c(2,2)); plot(as.numeric(simple), main="simple");
    noco = objects$opars$noco
    if(noco==3) { plot(as.numeric(full[,,1]), main="full"); plot(as.numeric(full[,,2]), main="full"); plot(as.numeric(full[,,3]), main="full") }
    if(noco!=3) plot(as.numeric(full[,,1]), main="full"); 
    browser();
    if(noco==3) { plot(as.numeric(simple)-as.numeric(full[,,1]), main="full"); plot(as.numeric(simple)-as.numeric(full[,,1]), main="full"); 
     plot(as.numeric(simple)-as.numeric(full[,,3]), main="full") }
    browser();
  }
}


make_nullcalls_na=function(hcalls_coll, nocl)
{
  hcalls_coll[which(hcalls_coll==(nocl+1))] = NA;
  return(hcalls_coll)
}

