my_rdirichlet <- function(n, a) {
   l <- length(a)
   x <- matrix(rgamma(l * n, a), ncol = l, byrow = TRUE)
   sm <- x %*% rep(1, l)
   x/as.vector(sm)
}

my_rmultinom <- function(p) length(p) - sum(runif(1) < cumsum(p)) + 1

mult_beta=function(alpha)
{
  check = sum(alpha)==1
  #if(!check) cat("Warning, sum of input PARAMETERS!=1 (", sum(alpha), ")\n", sep="")
  prod_gammas = prod(gamma(alpha))
  gamma_sum   = gamma(sum(alpha)) #Perhaps no needed in my case?
  mult_beta   = prod_gammas/gamma_sum
  return(mult_beta)
}

lmult_beta=function(alpha)
{
  check = sum(alpha)==1
  #if(!check) cat("Warning, sum of input PARAMETERS!=1 (", sum(alpha), ")\n", sep="")
  lprod_gammas = sum(lgamma(alpha))
  lgamma_sum   = lgamma(sum(alpha)) #Perhaps no needed in my case?
  lmult_beta   = lprod_gammas-lgamma_sum
  return(lmult_beta)
}

#This function could very easily take in a matrix and output a set of probabilities
d_dirichlet=function(data, alpha)
{
  if(check_dat(data, precision=5e-4)) return(0)
  mult_beta     = mult_beta(alpha)
  r_mult_beta   = 1/mult_beta
  data_to_alpha = data^(alpha-1)
  ddirichlet    = r_mult_beta*prod(data_to_alpha)
  return(ddirichlet)
}

ld_dirichlet=function(data, alpha)
{
  if(check_dat(data, precision=5e-4)) return(0)
  lmult_beta      = lmult_beta(alpha)
  lr_mult_beta    = -lmult_beta
  ldata_to_alpha  = (alpha-1)*log(data)
  lddirichlet     = lr_mult_beta + sum(ldata_to_alpha)
  return(lddirichlet)
}

check_dat=function(data, precision=1e-8, sum=NULL)
{
  if( is.null(sum)) check = (abs(sum(data)-1  )<precision)
  if(!is.null(sum)) check = (abs(sum(data)-sum)<precision)
  if(!check) {
    cat("Warning, sum of input DATA!=1 (", sum(data), ")\n", sep="")
    if(DEBUG>1.7 | abs(sum(data)-1)>0.05) browser();
    return(1)
  }
  return(0)
}

#This is my implementation of the Students-t distribution (wrapping around R's)
#This version aims to behave like dnorm, taking in a mean and standard deviation
#The t-dist is a N(0,1) divided by sqrt(chi squared/df)
my_dt=function(dat, mean, sd, df, log)
{
   dt          = dt((dat-mean)/sd, df=df, log=log)
   if( log) dt = dt-log(sd)
   if(!log) dt = dt/sd
   return(dt)
}
my_rt=function(mean, sd, df, nsamp=1)
{
   rt          = (rt(n=nsamp, df=df)*sd)+mean
   return(rt)
}





##Takes in two vectors, log(a) and log(b) and returns a vector log(a+b)
add_log=function(vec1, vec2)
{
   if(length(vec1)!=length(vec2)) cat("Error! Function add_log is written for 2 vectors of the same length.\n");
   biggest = vec1>vec2
   val  = rep(0, length(vec1))
   v1b  = vec1[which( biggest)]
   v2b  = vec2[which( biggest)]
   v1nb = vec1[which(!biggest)]
   v2nb = vec2[which(!biggest)]
   val[which( biggest)] = v1b  + log(1 + exp(v2b -v1b ))
   val[which(!biggest)] = v2nb + log(1 + exp(v1nb-v2nb))
   return(val)
}

##Takes in a vector, log(a) and returns a scalar log(sum(a))
add_log2=function(vec)
{
   biggest_val = max(vec)
   sum = biggest_val + log(sum(exp(vec-biggest_val)))
   return(sum)
}

add_log2mat=function(index, mat, MARGIN=1)
{
  if(MARGIN==1) vec = mat[index,] else vec = mat[,index]
  return(add_log2(vec=vec))
}

# sapply(1:dim(array)[1], add_log2array, array=array, MARGINS=c(1,1))
add_log2array=function(index, array, MARGINS=c(1,2))
{
  if(MARGINS[1]==1){ mat = array[index,,]; new_MARGIN = MARGINS[2];}
  if(MARGINS[1]==2){ mat = array[,index,]; new_MARGIN = MARGINS[2];}
  if(MARGINS[1]==3){ mat = array[,,index]; new_MARGIN = MARGINS[2];}
  
  add_log2mat(index, mat, MARGIN=new_MARGIN)
  sapply(1:dim(array)[MARGINS[1]], add_log2mat, mat=mat, MARGIN=MARGINS[2])
}

##Takes in two vectors, log(a) and log(b) and returns a vector log(a-b)
subtract_log=function(vec1, vec2)
{
   if(length(vec1)!=length(vec2)) cat("Error! Function add_log is written for 2 vectors of the same length.\n");
   biggest = vec1>vec2
   if(sum(biggest)<length(biggest)) {
     cat("Warning - only possible to subtract smaller from larger in log-space.\n");
     if(DEBUG>0.5) browser();
     if(DEBUG<0.5) stop(1);
   }
   val  = rep(0, length(vec1))
   v1b  = vec1[which( biggest)]
   v2b  = vec2[which( biggest)]
   v1nb = vec1[which(!biggest)]
   v2nb = vec2[which(!biggest)]
   val[which( biggest)] = v1b  + log(1 - exp(v2b -v1b ))
   val[which(!biggest)] = v2nb + log(exp(v1nb-v2nb) - 1)
   return(val)
}


ls.nonfun <- function(env=globalenv(), all.names=F) {
   obj.names <- ls(envir=env, all.names=all.names)
   is.fn <- sapply(obj.names, function(x) is.function(eval(as.name(x),envir=env)))
   return(obj.names[!is.fn])
}

#Give the size of objects that are not functions
# Alternative environments: ls.obsize(env=environment()), env = globalenv()
ls.obsize <- function(env = globalenv(), no_to_show=10) {
   non.functions <- ls.nonfun(env=env)
   show = min(no_to_show, length(non.functions))
   if(length(non.functions) > 0) {
     bytes <- sort(sapply(non.functions, function(s) object.size(get(s,env=env))), decreasing=TRUE)
     objnames = names(bytes)
     rv = paste("Size of ", objnames[1:show], " = ", signif(bytes[1:show] * 2^(-20), digits=3), "Mb", sep="")
     return(rv)
   }
   else return(non.functions)
}

##Memory check         #usage:    mem_summary(mem1=ls())
mem_summary=function(mem1)
{
  #mem1 = ls()
  mems = vector("numeric", length=length(mem1))  
  names(mems) = mem1
  i=1
  for(object in mem1)
  {
     #cat(object, "\t\tis using up\t", object.size(get(object))/1e6, "Mb of memory.\n");
     mems[i] = object.size(get(object))/1e6
     i=i+1
  }
  big_mem_users = order(mems,decreasing=T)[1:10]
  for(object in 1:length(big_mem_users))
  { 
     cat(mem1[big_mem_users[object]], "\t\tis using up\t", object.size(get(mem1[big_mem_users[object]]))/1e6, "Mb of memory.\n");
  }
}
  
###  This function converts the standard form parameters of a given distribution to the mean and variance ###
#mv = st_pars_to_mean_var(distribution="gamma"       ,  scale=1/20 ,  shape=20, verbose=T)
#mv = st_pars_to_mean_var(distribution="inverse_gamma", iscale=5, ishape= 500, verbose=T)
st_pars_to_mean_var=function(distribution="gamma", scale, shape, shape1, shape2, verbose=F, iscale, ishape)
{
  known_dists = c("gamma", "beta", "inverse_gamma")
  if(distribution=="gamma")
  {
    mean = scale  *shape
    var  = scale^2*shape
  }
  if(distribution=="beta")
  {
    mean = shape1/(shape1+shape2)
    var  = (shape1*shape2)/((shape1+shape2)^2 * (shape1+shape2+1))
  }
  if(distribution=="inverse_gamma")
  {
    check=(ishape>2); if(!check) {cat("Error. Can't calculate Inverse Gamma moments for ishape<=2.\n"); browser();}
    mean = iscale/(ishape+1)
    var  = iscale^2/((ishape-1)^2 * (ishape-2))
  }

  if(!(distribution %in% known_dists))
  {
    cat("Sorry, don't know how to convert for this distribution.\n")
    browser();
    return(c(NA, NA))
  }
  if(verbose) cat("Mean     = ", round(mean, 7), "\nVariance = ", round(var, 7), "\n");
  rv        = c( mean ,  var )
  names(rv) = c("mean", "var")
  return(rv)
}

###  This function converts the mean and variance of a given distribution to the standard parameters ###
# ss = mean_var_to_standard_pars(distribution="gamma"        , mean=0.02, var=0.003^2, verbose=T)
# ss = mean_var_to_standard_pars(distribution="beta"         ,  mean=0.5, var=0.035, verbose=T)
# ss = mean_var_to_standard_pars(distribution="inverse_gamma", mean=0.0002, var=(0.0002)^2, verbose=F)
mean_var_to_standard_pars=function(distribution="gamma", mean, var, verbose=F)
{
  known_dists = c("gamma", "beta", "inverse_gamma")
  if(distribution=="gamma")
  {
    scale = var/mean
    shape = mean/scale
    rv        = c( scale ,  shape )
    names(rv) = c("scale", "shape")
    if(verbose) cat("Scale = ", round(scale, 6), "\nShape = ", round(shape, 6), "\n");
  }
  if(distribution=="inverse_gamma")
  {
    ishape = (mean^2/var)+2
    iscale = mean*(ishape-1)
    rv        = c( iscale ,  ishape )
    names(rv) = c("iscale", "ishape")
    if(verbose) cat("iScale = ", round(iscale, 6), "\niShape = ", round(ishape, 6), "\n");
  }
  if(distribution=="beta")
  {
    shape1 = mean*(mean*(1-mean)/var -1)
    shape2 = (shape1/mean)-shape1
    rv        = c( shape1 ,  shape2 )
    names(rv) = c("shape1", "shape2")
    if(verbose) cat("shape1 = ", round(shape1, 6), "\nshape2 = ", round(shape2, 6), "\n");
  }

  if(!(distribution %in% known_dists))
  {
    cat("Sorry, don't know how to convert for this distribution.\n")
    browser();
    return(c(NA, NA))
  }
  return(rv)
}

normalise=function(vec, sum=NULL)
{
  if( is.null(sum)) rv = vec/sum(vec)
  if(!is.null(sum)) rv = vec/sum(vec)*sum
  return(rv)
}

if(0){
  mm2 = matrix(1:3, 1); colnames(mm2) = c("A", "B", "C"); rownames(mm2) = c("a");
  myas.matrix(apply(mm2, 2, normalise), rdim=dim(mm2), dimnames=dimnames(mm2))
  #Alternatively: 
  myapply(X=mm2, MARGIN=2, FUN=normalise)
}
myas.matrix=function(invec, rdim, dimnames=NULL)
{
  if( is.matrix(invec))  {
    rtmat           = invec
  } else                 {
    rtmat = matrix(invec, nrow=rdim[1], ncol=rdim[2])
    if(rdim[1]==1) colnames(rtmat) = names(invec)
    if(rdim[2]==1) rownames(rtmat) = names(invec)
  }
  if(!is.null(dimnames)) dimnames(rtmat) = dimnames
  return(rtmat)
}
# (!) added the ... arguments well after writing CNVCALL as a matter of style, if this function breaks this is probably why!
myapply=function(X, MARGIN, FUN, ...)
   return(myas.matrix(apply(X=X, MARGIN=MARGIN, FUN=FUN, ...), rdim=dim(X), dimnames=dimnames(X)))

#A wrapper for pmax that copes with length zero objects and retains the dimensions and names of 1xn and nx1 matrices.
mypmax=function(inmat, comparator)
{
   if(length(inmat)==0)  return(inmat)
   if(!is.matrix(inmat)) return(pmax(inmat, comparator))
   dn  = dimnames(inmat)
   dim = dim(inmat)
   pmv = pmax(inmat, comparator)
   if(!is.matrix(pmv)) return(matrix(pmv, nrow=dim[1], ncol=dim[2], dimnames=dn)) else return(pmv)
}
# make_imat(rownames=c("a","b"), colnames=c(3,6,8,9,1))
make_imat=function(rownames, colnames, rowtype=NULL, coltype=NULL)
{
  # mkel(5, rownames, colnames)
  firstimat = matrix(1:(length(rownames)*length(colnames)), nrow=length(rownames), ncol=length(colnames))
 #imat =  apply(X=firstimat, MARGIN=c(1,2), FUN=mkel, rownames=rownames, colnames=colnames)
  imat = matrix(lapply(X=firstimat, FUN=mkel, rownames=rownames, colnames=colnames, rowtype=rowtype, coltype=coltype), nrow=length(rownames))
  return(imat)
}
mkel = function(mat_index, rownames,colnames, rowtype=NULL, coltype=NULL) {
  col = ceiling(mat_index/length(rownames))
  row = mat_index - ((col-1)*length(rownames))
  rv = c(rownames[row],colnames[col])
  if(!is.null(rowtype) & !is.null(coltype)) names(rv) = c(rowtype,coltype)
  return(rv)
}

calls_cor=function(h_calls, cnv_types, nocl)
{
  rv        = mergeNocos(infList=h_calls, cnv_types=cnv_types)
  noco      = rv$noco
  cnv_types = rv$cnv_types

  this_cor=rep(-1, noco)
  for(coll in 1:noco)
  {
    mono_calls = range(h_calls[[coll]], na.rm=T)[1]==range(h_calls[[coll]], na.rm=T)[2]
    if( mono_calls & range(cnv_types)[1]!=range(cnv_types)[2]) this_cor[coll] = 0
    if( mono_calls & range(cnv_types)[1]==range(cnv_types)[2]) this_cor[coll] = 1
    if(nocl >1     & range(cnv_types)[1]==range(cnv_types)[2]) this_cor[coll] = 0
   #null_calls = which(is.na(h_calls[[coll]]))
   #if(length(null_calls)> 0) keep = (1:length(h_calls[[coll]]))[-null_calls]
   #if(length(null_calls)==0) keep = (1:length(h_calls[[coll]]))
    non_null_calls = which(!is.na(h_calls[[coll]]))
    if(length(non_null_calls)==0) this_cor[coll] = 0
    if(length(non_null_calls)> 0)
      if(!mono_calls & range(cnv_types)[1]!=range(cnv_types)[2]) this_cor[coll] = cor(h_calls[[coll]][non_null_calls], cnv_types[[coll]][non_null_calls])^2
    #cat("Correlation^2 of calls with simulated cnv_types in collection", coll, ":", this_cor, "\n");
  }
  return(this_cor)
}


# This funciton calculates correlations even when one vector is monomorphic or when there is NA data. 
# When both are monomorphic (and they agree) it returns 1, when only 1 or they don't is it returns 0.
# When there is NA data and na.rm=T then the correlation is calculated only on elements where both vectors are not NA.
# cor_with_mono(tvec=, ivec=, na.rm=T, DEBUG=0.3)
cor_with_mono=function(tvec, ivec, site=NULL, na.rm=T, DEBUG=0)  {
  napos = which(is.na(tvec) | is.na(ivec))
  if(length(napos)>0) {
    if(na.rm) {
      tvec  = tvec[-napos]
      ivec  = ivec[-napos]
    }
    else {
      cat("  Warning, can't calculate correlations with missing values (use na.rm=T?)\n");
      if(DEBUG>0.2) browser();
      return(NA)
    }
  } 
  tmono = sum(tvec==rep(tvec[1], length(tvec)))==length(tvec)
  imono = sum(ivec==rep(ivec[1], length(ivec)))==length(ivec)
  agree = (tvec[1]==ivec[1])
  if(tmono & imono & agree)
    { if(!is.null(site)) cat("Site", site, "seems to be a true monomorphic.\n"); return(1) }
  if(!tmono & !imono)
  {
    correlation = cor(tvec, ivec)
    return(correlation)
  }
  return(0)
}


## Calculating MAF and HWE p-values given posterior calls for CNV genotypes.
##
## Damjan Vukcevic
##
## Created: 20 May 2009
## Updated: 20 May 2009

## Input data is an (n x k) matrix, where n is the number of individuals and k
## the number of components (i.e. genotype classes) for a given CNV.  Each row
## gives a posterior probability of each component.  These don't need to sum
## to 1, the remaining probability mass will be attirbuted to a NULL call.
##
## Both the MAF and HWE calculations will assume the CNV is biallelic when k is
## 1, 2, or 3, and will perform the same calcluation as would be applicable for
## SNPs.  When k is greater, a value of NA is returned.


# =============================================================================
#
# Main functions.

# Calculates the MAF, assuming the CNV is biallelic.  Uses _expected_ genotype
# counts.
calc.maf <- function(cnv.calls)
{
    # Number of components.
    k <- ncol(cnv.calls)
    # Give up if too many components.
    if (k >= 4)
        return(NA)
    # If monomorphic, just return 0.
    if (k == 1)
        return(0)
    # Expected genotype frequencies, converted to 3 components if necessary
    # (i.e. k is now effectively equal to 3).
    geno.freqs <- expected.genotype.freqs(cnv.calls)
    # Calculate the MAF, assuming CNV is biallelic.
    m <- 0.5 * geno.freqs[2] + min(geno.freqs[c(1,3)])
    return(m)
}

# Calculates the p-value for a HWE test.  Assumes the CNV is biallelic.  Uses
# _expected_ genotype counts.  Uses the asymptotic chi^2 test (not the exact
# test).
calc.hwe.pval.asymp <- function(cnv.calls)
{
    # Number of components.
    k <- ncol(cnv.calls)
    # Give up if too many components.
    if (k >= 4)
        return(NA)
    # If monomorphic, the CNV automatically satisfies HWE, so just return a
    # p-value of 1.
    if (k == 1)
        return(1)
    # Expected genotype frequencies, converted to 3 components if necessary
    # (i.e. k is now effectively equal to 3).
    geno.freqs <- expected.genotype.freqs(cnv.calls)
    # Calculate the allele frequency.
    af <- drop(c(0, 0.5, 1) %*% geno.freqs)
    # Calculate the effective sample size.
    n <- sum(cnv.calls)
    # Calculate the genotype counts expected under HWE.
    geno.freqs.exp <- n * c((1-af)^2, 2*af*(1-af), af^2)
    # The observed genotype counts.
    geno.freqs.obs <- n * geno.freqs
    # The Pearson chi^2 statistic comparing the two count distributions.
    chi2 <- sum((geno.freqs.obs - geno.freqs.exp)^2 / geno.freqs.exp)
    # The p-value corresponding the the test statistic.
    pval <- pchisq(chi2, df=1, lower.tail=F)
    return(pval)
}

# Calculates the p-value for a HWE test.  Assumes the CNV is biallelic.  Uses
# _rounded_ _expected_ genotype counts.  Uses the exact test.
# Based on code from Bryan Howie.
calc.hwe.pval.exact <- function(cnv.calls)
{
    # Number of components.
    k <- ncol(cnv.calls)
    # Give up if too many components.
    if (k >= 4)
        return(NA)
    # If monomorphic, the CNV automatically satisfies HWE, so just return a
    # p-value of 1.
    if (k == 1)
        return(1)
    # Expected genotype frequencies, converted to 3 components if necessary
    # (i.e. k is now effectively equal to 3).
    geno.freqs <- expected.genotype.freqs(cnv.calls)
    # Calculate the effective sample size.
    n <- sum(cnv.calls)
    # _Rounded_ genotype counts.
    geno.counts.rounded <- round(n * geno.freqs)
    # Count genotypes and alleles.
    n00 <- geno.counts.rounded[1]
    n01 <- geno.counts.rounded[2]
    n11 <- geno.counts.rounded[3]
    n0  <- 2*n00 + n01
    n1  <- 2*n11 + n01
    n   <-   n00 + n01 + n11
    # Create lookup table of probabilities of genotype counts consistent with
    # observed allele counts.
    x01 <- seq(n01 %% 2, min(n0,n1), 2)
    x00 <- (min(n0,n1) - x01) %/% 2   # arbitrarily code rare allele as '0' for table
    x11 <- (max(n0,n1) - x01) %/% 2
    dist <- data.frame(n00=x00, n01=x01, n11=x11,
                       prob=cond.prob.het(x00, x01, x11))
    dist <- dist[order(dist$prob),]
    dist$cum.prob <- cumsum(dist$prob)
    # Determine p-value by looking up the observe genotype counts in the above
    # table.
    rare.hom <- min(n00,n11)
    comm.hom <- max(n00,n11)
    pval <- dist$cum.prob[dist$n00 == rare.hom &
                          dist$n01 == n01      &
                          dist$n11 == comm.hom]
    return(pval)
}


# =============================================================================
#
# Helper functions.

# Calculated expected genotype frequencies given a matrix of calls.  Be default
# convert into a 3-component set of frequencies, emulating a biallelic SNP.
expected.genotype.freqs <- function(cnv.calls, create.3class=TRUE)
{
    # Expected frequencies (renormalised expected genotype counts).
    expected.freqs <- colSums(cnv.calls) / sum(cnv.calls)
    # Return this vector if that's all that is required.
    if (!create.3class)
        return(expected.freqs)
    # Otherwise, continue with creating SNP-like counts...
    # Number of components.
    k <- length(expected.freqs)
    # Determine how to proceed depending on the number of components.
    if (k == 3) {
        # If 3 components, we are done.
        return(expected.freqs)
    } else if (k == 2) {
        # If 2 components, add a zero-count third component, assuming the less
        # frequent component is the heterozygote.
        het.index <- which.min(expected.freqs)
        if (het.index == 1)
            expected.freqs <- c(0, expected.freqs)
        else if (het.index == 2)
            expected.freqs <- c(expected.freqs, 0)
        else
            stop('Error in genotype frequency calculation: het.index is an unrecognised value.')
        return(expected.freqs)
    } else if (k >= 4 || k <= 0) {
        # Give up if the number of components is not compatible with being
        # biallelic.
        return(NA)
    } else {
        # Should never reach this point.
        stop('Error in genotype frequency calculation: k is an recognised value.')
    }
    # Should never reach this point.
    stop('Error in genotype frequency calculation.')
}

# Conditional probability of the heterozygote count given the allele counts.
# Based on code from Bryan Howie.
cond.prob.het <- function(n00, n01, n11) {
    # Total and allele counts.
    n  <-   n00 + n01 + n11
    n0 <- 2*n00 + n01
    n1 <- 2*n11 + n01
    # log gamma function (i.e. log factorial)
    l <- function(x) lgamma(x+1)
    # Conditional probability of the heterozygote count given the allele
    # counts.
    exp((l(  n) - l(n00) - l(n01) - l(n11)) -
        (l(2*n) - l(n0)  - l(n1)) + 
        n01 * log(2))
}

# This function expects two tables to compare. The columns of each table are: name, category, start end
# Two intervals [start1, end1] and [start2, end2] can only match if they are in the same category (eg. chromsoome)
# The function returns a set of overlapping regions depending on the stringency specified
matchLocations=function(set1, set2, stringency="anyOverlap", DEBUG=1, name1="name1", name2="name2", decentMatchMin=1e3, pause=F)
{
  #For two sets of starts and ends: (s1, e1) and (s2, e2) the following is necessary and sufficient for an overlap:
  # ((s1<=s2) and (e1>=s2)) or ((s1>=s2) and (e2>=s1))

  categories    = unique(intersect(set1[,"category"], set2[,"category"]))
  decentMatches = as.list(rep(NA, length(categories))); names(decentMatches) = categories
  if(stringency=="anyOverlap")
  {
    for(category in categories)
    {
      cat("  On category: ", category, "\n");
      thisCatSubset1 = which(set1[,"category"]==category)
      thisCatSubset2 = which(set2[,"category"]==category)
      i1 = set1[thisCatSubset1, c("start","end")]; rownames(i1) = set1[thisCatSubset1, "name"]
      i2 = set2[thisCatSubset2, c("start","end")]; rownames(i2) = set2[thisCatSubset2, "name"]
      #intervals  = list(i1, i2); names(intervals) = c(name1, name2)
      #longSet  = which.max(sapply(intervals, nrow))
      #shortSet = which.min(sapply(intervals, nrow))
      #intervals1 = intervals[[longSet]]
      #intervals2 = intervals[[shortSet]]
      overlapMat    = findIntervalOverlaps(intervals1=i1, intervals2=i2, DEBUG=DEBUG, name1=name1, name2=name2) 
      best_matches  = apply(overlapMat, MARGIN=1, which.max)
      matchLensBest = apply(overlapMat, MARGIN=1, max)
      matchLens     = pmax(decentMatchMin, matchLensBest)
      tmp           = which(overlapMat>=matchLens, arr.ind=T) 
      tmp2          = cbind(rownames(overlapMat)[tmp[,1]], colnames(overlapMat)[tmp[,2]])
      decentMatches[[category]] = cbind(tmp, tmp2)
      if(pause) browser()
    }
  } 
  return(decentMatches)
}

# find all the overlaps between two sets of intervals
# is1 = matrix(1:6, ncol=2); colnames(is1)=c("start", "end"); is2 = matrix(c(0,5, -2, -1, 0, 2, 4.5, 7, 7,8), ncol=2, byrow=T); colnames(is2)=c("start", "end"); findIntervalOverlaps(intervals1=is1, intervals2=is2, DEBUG=2)
findIntervalOverlaps=function(intervals1, intervals2, DEBUG=1, name1="name1", name2="name2")
{
  allOverlaps = apply(intervals1, MARGIN=1, checkIntervalsVsIntervalOverlap, vectorIntervals=intervals2, DEBUG=DEBUG, name=name2)
  if(is.null(rownames(intervals1))) colnames(allOverlaps) = paste(name1, "_", 1:nrow(intervals1), sep="") else colnames(allOverlaps) = rownames(intervals1)
  return(allOverlaps)
}

# Check for the overlap of a set of intervals with a specific interval: vectorIntervals=(s1, e1) and singleInterval=(s, e).
# is1 = matrix(c(1,2,3,4), 2); colnames(is1)=c("start", "end"); si = c("start"=0, "end"=1); checkIntervalsVsIntervalOverlap(singleInterval = si, vectorIntervals = is1)
checkIntervalsVsIntervalOverlap=function(singleInterval, vectorIntervals, DEBUG=1, name="vec")
{
  overlaps        = apply(vectorIntervals, MARGIN=1, checkIntervalOverlap, i2=singleInterval, DEBUG=DEBUG)
  if(is.null(rownames(vectorIntervals))) names(overlaps) = paste(name, "_", 1:nrow(vectorIntervals), sep="") else names(overlaps) = rownames(vectorIntervals)
  return(overlaps)
}

# Check for the overlap of two intervals on the real line defined by: i1=(s1, e1) and i2=(s2, e2).
checkIntervalOverlap=function(i1, i2, DEBUG=1, verbose=1, getLen=TRUE)
{
  s1=i1[1]; e1=i1[2]; s2=i2[1]; e2=i2[2]
  checkOrder = (s1<=e1)*(s2<=e2)
  if(!checkOrder) {cat("  Warning! Haven't supplied consistent starts and ends.\n"); if(DEBUG>1) browser();}
  if(!getLen) {
    type1 = (s1<=s2) & (e1>=s2)
    type2 = (s1>=s2) & (e2>=s1)
    overlap = type1 | type2
    if(verbose>=2) {
      cat("s1 = ", s1, ", e1 = ", e1, ", s2 = ", s2, ", e2 = ", e2, "\n", sep="")
      cat("type1 = ", type1, ", type2 = ", type2, "\n", sep="")
    }
    if(overlap) overlapLen=1 else overlapLen=0
  } else {
    overlapLen = max(0, min(e2,e1)-max(s1,s2))
  }
  return(overlapLen)
}



visualise_tDsitributions=function(nsamp=1e5, numProbes=10)
{
  x3 = my_rt(mean=0, sd=1, df=3, nsamp=nsamp)
  x8 = my_rt(mean=0, sd=1, df=8, nsamp=nsamp)
  xn = my_rt(mean=0, sd=1, df=1000, nsamp=nsamp)

  x3mat = matrix(my_rt(mean=0, sd=sqrt(numProbes), df=3   , nsamp=nsamp), ncol=numProbes)
  x8mat = matrix(my_rt(mean=0, sd=sqrt(numProbes), df=8   , nsamp=nsamp), ncol=numProbes)
  xnmat = matrix(my_rt(mean=0, sd=sqrt(numProbes), df=1000, nsamp=nsamp), ncol=numProbes)

  plot(density(x3), col="grey", ylim=c(0, 0.5), xlim=c(-5, 5))
  lines(density(x8), col="green")
  lines(density(xn), col="red")

  rM3 = rowMeans(x3mat)
  lines(density(rM3), col="blue")
  lines(density(rowMeans(x8mat)), col="orange")
  lines(density(rowMeans(xnmat)), col="purple")
  browser()
}
# visualise_tDsitributions(nsamp=2e5)



