
#This function makes the structure that the program expects so that it can be used with an arbitrary data_matrix
# Eg. to test it run:    source("cnv_caller.R"); mat = matrix(rnorm(1000, 0, 1), nrow=10, ncol=100); analyse_data(dat_matrix=mat)
#
# TODO: need to pass process_data an argument that specifies the duplicates
# (or more generally, individuals with repeated measurements).  This will come
# from arguments(?).
make_data_and_details=function(dat_matrix=NULL, data_array=NULL, arguments=set_arguments(write_results=F), wdat=0.1, DEBUG=1.71, indAttributes=NULL)
{
  
  if(is.null(data_array)) {
    if(ncol(dat_matrix)>nrow(dat_matrix)) {
      cat("  Warning - assume that columns are probes. \nSeem to have more columns than rows (individuals).\n");
      if(DEBUG>1.5) browser() else {
        cat("  Transposing!\n");
        dat_matrix = t(dat_matrix)    
      }
    }
    data_array   = matrix(list(""), nrow=1, ncol=1)
    data_array[1,1][[1]] = dat_matrix    
    names(data_array)    = makeDefaultCollnames(noco=ncol(data_array))
    colnames(data_array) = names(data_array)
  }
  no_inds_vec = unlist(apply(data_array, 2, get_no_inds))
  duplicates  = myload(arguments$duplicates_fn)
  fdat        = process_data(data_array,
                             no_inds          = sum(no_inds_vec),
                             keep_range       = FALSE,
                             wdat             = wdat,
                             DEBUG            = DEBUG,
                             return_ungridded = TRUE,
                             grid_density     = arguments$grid_density,
                             duplicates       = duplicates,
                             probe_scaling    = arguments$probe_scaling,
                             probe_summary    = arguments$probe_summary,
                             verbose          = arguments$verbose)
  gdat         = fdat$gridded
  grid_breaks  = fdat$grid_breaks
  true_d_range = fdat$true_d_range
  probe_scale  = fdat$probe_scale  # the option used for probe scaling
  noco         = ncol(data_array)
  ung_dat      = fdat$ungridded #This data is a LIST (As collections may have different sizes)
  total_sim    = NULL
  cnv_name     = NULL
  cnv_index    = 1;
  used_ogt_sample_names = NULL
  data_and_details = list("gdat"                  = gdat,
                          "data_array"            = data_array,
                          "ung_dat"               = ung_dat,
                          "noco"                  = noco,
                          "total_sim"             = total_sim,
                          "cnv_name"              = cnv_name,
                          "used_ogt_sample_names" = used_ogt_sample_names,
                          "grid_breaks"           = grid_breaks,
                          "cnv_index"             = cnv_index,
                          "true_d_range"          = true_d_range,
                          "indAttributes"         = indAttributes,
                          "probe_scale"           = probe_scale)
}



identify_extreme_inds=function(dm, extension=0.5, q_cutoff = 3e-3, DEBUG)
{
  if(!exists("DEBUG")) {cat("Warning - no instance of DEBUG passed (r.e.i.)\n"); DEBUG=1}
  check       = sum(colnames(dm)[1:3] == c("Chr", "StartCoord", "ProbeIndex"))==3
  if(DEBUG>1 & !check) {cat("Potential problem with data matrix processing.\n"); browser();}
  dmeans      = colMeans(dm[,-(1:3), drop=F])
  extreme_pts = quantile(dmeans, prob=c(q_cutoff, 1-q_cutoff))
  width       = max(extreme_pts)-min(extreme_pts)
  cut_off     = c(extreme_pts[1]-(width*extension), extreme_pts[2]+(width*extension))
  remove_inds = c(which(dmeans<cut_off[1]), which(dmeans>cut_off[2]))
 #if(length(remove_inds)>0 & length(remove_inds) < 6) dm = dm[,-remove_inds, drop=F]
  return(remove_inds)
}


#sapply((dimnames(dat_matrix)[[2]])[1:4], remove_x)
remove_x=function(string)
{
  check = (substr(string, 1, 1)=="X")
  if(check) rv = substr(string, 2, nchar(string)) else rv = string
  return(rv)
}


